package application;
	
import java.io.File;
import java.io.IOException;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.xml.sax.SAXException;

import application.world.World;
import application.world.WorldActorOutOfGridException;
import application.world.WorldActorUnsupportedOperationException;
import application.world.WorldNotInStateToPerformSimulationStep;
import application.world.WorldState;
import javafx.application.Application;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.TextArea;
import javafx.scene.layout.FlowPane;
import javafx.scene.layout.GridPane;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.stage.Stage;

/**
 * Main JavaFX Application class.
 * It contains entry point (start() method) for the Application as well as it encapsulates GUI Nodes. 
 * updated during the simulations.
 * 
 * @author Pavel Bory 
 */
public class Main extends Application {
	private static final Logger logger = LogManager.getLogger(Main.class);
	
	// Relative path to a xml file with default world configuration loaded at application start
	public static final String DefaultWorldFilePath = "bin/worlds/default/world.xml";
	// Relative path used by FileChoosers when opening / saving files
	public static final String FileChooserInitialPath = "bin";
	
	// FXML Controls Id Selectors
	public static final String MainCanvasIdSelector = "#mainCanvas";
	public static final String DebugInfoTextAreaIdSelector = "#debugDialogMainTextArea";
	public static final String ActorsOutputTextAreaIdSelector = "#textAreaActorsOutput";
	public static final String BtnSimulationRunIdSelector = "#btnSimulationRun";
	
	// Controls which are updated during the simulation
	private static World world;
	private static Canvas mainCanvas;
	private static Stage primaryStage; 
	private static Stage debugInfoStage;
	private static TextArea debugInfoTextArea;
	private static TextArea actorsOutputTextArea;
	private static Button btnSimulationRun;
	
	public static void main(String[] args) {
		launch(args);
	}
	
	// JavaFX Application Start Method
	@Override
	public void start(Stage primaryStage) {
		try {
			logger.info("Application starting");
			GridPane root = FXMLLoader.load(getClass().getResource("MainPane.fxml"));
			Scene scene = new Scene(root,1280,768);
			scene.getStylesheets().add(getClass().getResource("application.css").toExternalForm());
			primaryStage.setScene(scene);
			primaryStage.show();
			primaryStage.setTitle("Pavel Bory - Diplomová práce ČZU, 2017");
			
			Main.mainCanvas = (Canvas)root.lookup(MainCanvasIdSelector);
			Main.primaryStage = primaryStage;
			Main.actorsOutputTextArea = (TextArea)root.lookup(ActorsOutputTextAreaIdSelector);
			Main.btnSimulationRun = (Button)root.lookup(BtnSimulationRunIdSelector);
			Main.world = initializeWorld(mainCanvas,null);
			
			initializeDebugDialog();
			logger.info("First World initialization on application startup successfully completed. Displaying Primary Stage.");
		} catch(Exception e) {
			logger.error("Application failed to start: ", e);
		}
	}
	
	/**
	 * Initializes Simulation World based on the configuration passed by worldFile.
	 * World Simulation will use canvas to draw a World's Grid.
	 * 
	 * @author Pavel Bory
	 * @param canvas Canvas to draw Simulation World's Grid
	 * @param worldFile File with the World initial configuration
	 * @return Instance of the Simulation World based on the given Configuration File
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 */
	private static World initializeWorld(Canvas canvas, File worldFile) throws SAXException, IOException, ParserConfigurationException{
		if(worldFile == null) {
			worldFile = new File(DefaultWorldFilePath);
		}
		logger.debug("Initializing World: {}", worldFile.getAbsolutePath());
		World world = new World(canvas,worldFile);
		logger.debug("Initializing World: {} COMPLETED", worldFile.getAbsoluteFile());
		return world;
	}
	
	/**
	 * Initializes the Debug Dialog component and stores references to it's controls updated during the Simulation.
	 * @author Pavel Bory
	 */
	private void initializeDebugDialog(){
		try {
			logger.info("Initializing Debug Info Dialog");
			FlowPane debugInfoPane = FXMLLoader.load(getClass().getResource("DebugInfoDialog.fxml"));
			Scene dialogScene = new Scene(debugInfoPane);
			dialogScene.getStylesheets().add(getClass().getResource("application.css").toExternalForm());
			Main.debugInfoTextArea = (TextArea)debugInfoPane.lookup(DebugInfoTextAreaIdSelector);
			Main.debugInfoStage = new Stage();
			Main.debugInfoStage.setTitle("DEBUG INFO");
			Main.debugInfoStage.setScene(dialogScene);
			logger.info("Debug Info Dialog initialized");
		} catch (IOException e) {
			logger.error("Failed to initialize Debu Info Dialog: ", e);
		}
	}
	
	/**
	 * Switches the text of Button Run / Pause Simulation based on actual World state.
	 * @author Pavel Bory
	 */
	public static void updateSimulationRunButton(){
		if(Main.world.getState() == WorldState.RUNNING){
			logger.debug("Updating Simulation Run Button to PAUSE");
			btnSimulationRun.setText("Pause");
		}else{
			logger.debug("Updating Simulation Run Button to RUN");
			btnSimulationRun.setText("Run");
		}
	}
	
	/**
	 * Displays Alert Error Type Dialog
	 * @param errText Text describing the error
	 */
	private static void displayErrorDialog(String errText){
		Alert alert = new Alert(AlertType.ERROR);
		alert.setContentText(errText);
		alert.show();
	}
	
	/**
	 * Handles event - Load code entered by the user to the Ruby Actor instance
	 * @param code Ruby Source code to be loaded to the Ruby Actor instance.
	 * @author Pavel Bory
	 */
	public static void loadCodeForActor(String code){
		logger.debug("Loading code for the Actor");
		logger.trace(code);
		world.loadCodeForActor(code);
		logger.debug("Loading code for the Actor completed");
		
		Alert alert = new Alert(AlertType.INFORMATION);
		alert.setContentText("Code for Actor has been loaded. You can now run the Simulation.");
		alert.show();
	}
	
	/**
	 * Handles event - Perform one single step of the Simulation
	 * @author Pavel Bory
	 */
	public static void step(){
		logger.info("Performing single step of the simulation");
		try{
			world.step(true);
		}catch(WorldActorOutOfGridException e){
			logger.trace("Actor code caused excpetion: ",e);
			HandleWorldSimulationException(e);
		} catch (WorldActorUnsupportedOperationException e) {
			logger.trace("Actor code caused excpetion: ",e);
			HandleWorldSimulationException(e);
		} catch (WorldNotInStateToPerformSimulationStep e) {
			logger.trace("Actor code caused excpetion: ",e);
			HandleWorldSimulationException(e);
		}
		updateSimulationRunButton();
		logger.info("Performing single step of the simulation - COMPLETED");
	}
	
	/**
	 * Handles event - Reset the Simulation to it's initial state
	 * 
	 * @author Pavel Bory
	 */
	public static void reset(){
		logger.info("Reseting Simulation");
		world.reset();
		updateSimulationRunButton();
		logger.info("Reseting Simulation - COMPLETED");
	}
	
	/**
	 * Handles event - Run the Simulation continuously
	 * 
	 * @author Pavel Bory
	 */
	public static void run(){
		logger.info("Starting simulation in continuous mode");
		world.run();
		updateSimulationRunButton();
		logger.info("Starting simulation in continuous mode - COMPLETED");
	}
	
	/**
	 * Displays and brings to front the Debug Info Dialog
	 * 
	 * @author Pavel Bory
	 */
	public static void showDebugInfoDialog(){
		logger.info("Displaying the Debug Info Dialog");
		debugInfoStage.show();
		debugInfoStage.toFront();
		logger.info("Displaying the Debug Info Dialog - COMPLETED");
	}
	
	/**
	 * Sets the given text to the Debug Info Dialog
	 * 
	 * @param text Text to be displayed.
	 */
	public static void setDebugDialogText(String text){
		logger.trace("Text being displayed in the Debug Info Dialog: {}",text);
		if(debugInfoTextArea != null) {
			debugInfoTextArea.setText(text);
		}else{
			logger.trace("debugInfoTextArea not initialized yet");
		}
	}
	
	/**
	 * Lets user select a configuration file from which the Simulation World should be loaded.
	 * 
	 * @author Pavel Bory
	 */
	public static void loadWorldFromFile() {
		FileChooser fileChooser = new FileChooser();
		fileChooser.setTitle("Select World xml file");
		fileChooser.setInitialDirectory(new File(FileChooserInitialPath));
		fileChooser.getExtensionFilters().add(new ExtensionFilter("XML Documents", "*.xml"));
		File worldFile = fileChooser.showOpenDialog(primaryStage);
		if(worldFile != null){
			try {
				world = initializeWorld(mainCanvas,worldFile);
			} catch (Exception e) {
				logger.error("loadWorldFromFile failed: ",e);
			}
		}
	}
	
	/**
	 * Handles event - Clear actors output text area.
	 * 
	 * @author Pavel Bory
	 */
	public static void clearActorsOutput(){
		logger.trace("Clearing Actor Output Text Area");
		actorsOutputTextArea.setText("");
	}
	
	/**
	 * Appends given text to the actors output text area.
	 * 
	 * @param text
	 */
	public static void appendActorsOutput(String text){
		logger.trace("Appending text to Actor Output Text Area: {}",text);
		actorsOutputTextArea.appendText(text);
	}
	
	public static void HandleWorldSimulationException(WorldActorOutOfGridException e){
		displayErrorDialog(e.getMessage());
	}
	
	public static void HandleWorldSimulationException(WorldNotInStateToPerformSimulationStep e){
		displayErrorDialog(e.getMessage());
	}
	
	public static void HandleWorldSimulationException(WorldActorUnsupportedOperationException e){
		displayErrorDialog(e.getMessage());
	}
}
