package application.world;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import application.Main;
import application.world.ruby.JRubyContainer;
import application.world.ruby.RubyActorActionType;
import application.worldgraphics.WorldGrid;
import javafx.scene.canvas.Canvas;

/**
 * The entire World of the Simulation. Contains all items of the Simulation: WorldGrid, WorldSettings,Actor,JRubyContainer,...
 * 
 * This is the primary class to maintain and interact with the simulation entities.
 * 
 * @author Pavel Bory
 *
 */
public class World {
	private static final Logger logger = LogManager.getLogger(World.class);
	
	// Count of the World Grid columns
	public static final int WORLD_COLUMNS_COUNT = 10;
	// Count of the World Grid rows
	public static final int WORLD_ROWS_COUNT = 10;
	// Delay in ms before the Simulation starts when user "fires" the event to start the Simulation
	public static final int TIMER_DELAY = 0;
	// Period determining time between the Simulation steps
	public static final int TIMER_PERIOD = 1500;
	
	// Abstraction of the World Grid that maintains it's graphical not functional state
	private WorldGrid worldGrid;
	// Settings of the Simulation World. Based on these settings the Simulation World can be reset to INITIAL state
	private WorldSettings worldSettings;
	// Java synchronized Actor that holds actual state of the Actor in the Simulation
	private Actor actor;
	// Container where the JRuby code resides and through which it can be invoked
	private JRubyContainer jrubyContainer;
	// Actual state of the Simulation World
	private WorldState state;
	// Auxiliary variable that holds reference to the instance of Timer which periodicaly invokes task to perform one step of the Simulation 
	private Timer timer;
	
	/**
	 * Create new instance of the Simulation World in INITIAL state with given canvas, on which the World Grid should be drawn
	 * and given the file with World configuration.
	 * 
	 * @param canvas Canvas on which the Worlds graphics representation should be drawn.
	 * @param worldFile File with the World configuration.
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * 
	 * @author Pavel Bory
	 */
	public World(Canvas canvas, File worldFile) throws SAXException, IOException, ParserConfigurationException{
		logger.info("Creating new World instance");
		worldGrid = new WorldGrid(canvas);
		logger.debug("World Grid created");
		worldSettings = WorldSettings.load(worldFile);
		logger.debug("World Settings loaded from file");
		jrubyContainer = new JRubyContainer();
		logger.debug("JRubyContainer initialized");
		setInitialWorldState();
		logger.debug("World set to it's INITIAL state");
		drawCurrentState();
		logger.debug("Worlds graphical representation has been drawn on the canvas");
		setState(WorldState.INITIAL);
		updateDebugInfo();
		logger.info("Creating new World instance - Successfuly completed");
	}
	
	/**
	 * Loads the given source code to the Ruby Actor using the JRubyContainer
	 * @param code
	 * 
	 * @author Pavel Bory
	 */
	public void loadCodeForActor(String code){
		logger.info("Loading source code for the Actor");
		logger.trace("Source code: {}",code);
		jrubyContainer.loadCodeForActor(actor,code);
		setState(WorldState.READY);
		reset();
		updateDebugInfo();
		logger.info("Loading source code for the Actor - completed");
	}
	
	/**
	 * Resets the World State to initial state based on the World Settings.
	 * Actually sets initial coordinates to the Actor.
	 * 
	 * @author Pavel Bory
	 */
	public void setInitialWorldState(){
		actor = new Actor(worldSettings.getActorInitialPositionX(),worldSettings.getActorInitialPositionY());
		updateDebugInfo();
	}
	
	/**
	 * Performs one step of the Simulation, that consists of the following substeps:
	 * 		- Check whether the World is in correct state to perform next step (for example that it is not TERMINATED)
	 * 		- Call Ruby Actor makeAction code through JRubyContainer and save what Ruby Actor want's to do in this step of the Simulation
	 * 		- Gather Actor's text output (if any)
	 * 		- Based on the Actor's makeAction result perform requested action (if possible)
	 * 		- Re-draw World Grid after the step of the Simulation
	 * 		- Update Debug Info Dialog
	 * 
	 * @param singleStep Determines whether the method is called within the Continuous Simulation Run Mode or not
	 * @throws WorldActorOutOfGridException
	 * @throws WorldActorUnsupportedOperationException
	 * @throws WorldNotInStateToPerformSimulationStep
	 * 
	 * @author Pavel Bory
	 */
	public void step(boolean singleStep) throws WorldActorOutOfGridException, WorldActorUnsupportedOperationException, 
												WorldNotInStateToPerformSimulationStep
	{
		logger.info("Performing step. Single step: {}",singleStep);
		// Check if the WorldState enables to perform Actor's action
		if(singleStep){
			if(getState() == WorldState.TERMINATED || getState() == WorldState.INITIAL){
				logger.info("World in state: {} can't perform next step",getState());
				throw new WorldNotInStateToPerformSimulationStep(String.format("World is in state: %s, so it cannot perform step of the Simultion",getState()));
			}else{
				setState(WorldState.PAUSED);	
			}
		}else{
			if(getState() != WorldState.RUNNING){
				logger.debug("World in state {} can't perform a non single step. To make continuos steps it is expected that world is in state RUNNING",getState());
				return;
			}
		}
		
		// Call Ruby Actor's method that performs user programmed function and store Actor's text output (if any) and append it to the GUI 
		RubyActorActionType actionType = jrubyContainer.makeActorAction(actor,getActualItemInputForActor(actor));
		String actorsOutputText = jrubyContainer.gatherOutputText();
		if(actorsOutputText != null && !actorsOutputText.isEmpty()){
			// Simplified solution of newline symbols conversion between Java and Ruby Strings. __NEWLINE__ is internally considered to represent
			// newline symbol that comes from the Ruby Actor. It should be handled properly in next versions of this application and robust
			// special String symbols conversion mechanism should be implemented in the JRubyContainer.
			actorsOutputText = actorsOutputText.replaceAll("__NEWLINE__", "\n");
			Main.appendActorsOutput(actorsOutputText);
		}
		
		// Perform requested action (if possible)
		logger.debug("Action type: {}",actionType);
		switch (actionType) {
			case MOVE_RIGHT:
				if(actor.getPositionX() >= WORLD_COLUMNS_COUNT - 1){
					setState(WorldState.TERMINATED);
					throw new WorldActorOutOfGridException("The actor went of of the grid on the x-axis");
				}else{
					actor.setPositionX(actor.getPositionX() + 1);
				}
				break;
			case MOVE_LEFT:
				if(actor.getPositionX() <= 0){
					setState(WorldState.TERMINATED);
					throw new WorldActorOutOfGridException("The actor went of of the grid on the x-axis");
				}else{
					actor.setPositionX(actor.getPositionX() - 1);
				}
				break;
			case MOVE_UP:
				if(actor.getPositionY() <= 0){
					setState(WorldState.TERMINATED);
					throw new WorldActorOutOfGridException("The actor went of of the grid on the y-axis");
				}else{
					actor.setPositionY(actor.getPositionY() - 1);
				}
				break;
			case MOVE_DOWN:
				if(actor.getPositionY() >= WORLD_ROWS_COUNT - 1){
					actor.setPositionY(actor.getPositionY() + 1);
				}else{
					setState(WorldState.TERMINATED);
					throw new WorldActorOutOfGridException("The actor went of of the grid on the y-axis");
				}
				break;
			case TERMINATE:
				setState(WorldState.TERMINATED);
				break;
		}
		drawCurrentState();
		updateDebugInfo();
		logger.info("Performing step. Single step: {} - COMPLETED",singleStep);
	}
	
	/**
	 * Calls appropriate World Grid drawing methods to draw the current graphical representation of the World.
	 * 
	 * @author Pavel Bory
	 */
	public void drawCurrentState(){
		logger.debug("Drawing Current State");
		worldGrid.drawGrid(worldSettings.getGridCellColors());
		logger.debug("Draw World Grid - COMPLETED");
		worldGrid.drawItems(worldSettings.getGridWorldItems());
		logger.debug("Draw World Grid Items - COMPLETED");
		worldGrid.drawActor(actor.getPositionX(),actor.getPositionY(),worldSettings.getActorImg());
		logger.debug("Draw Actor - COMPLETED");
	}
	
	/**
	 * Resets the World to initial state based on the World Settings.
	 * Reset is not performed when the world state is INITIAL because for INITIAL state it is unnecessary.
	 * 
	 * @author Pavel Bory
	 */
	public void reset(){
		if(getState() != WorldState.INITIAL){
			logger.debug("Reseting world");
			setState(WorldState.READY);
			setInitialWorldState();
			drawCurrentState();
			updateDebugInfo();
			logger.debug("Reseting world - COMPLETED");
		}else{
			logger.debug("World can't be reseted now, because the World is in INITIAL state.");
		}
	}

	/**
	 * Starts the continuous Simulation, so that the steps of the simulation are performed in a given period (based on World.TIMER_PERIOD).
	 * Start can be delayed (based on World.TIMER_DELAY)
	 * 
	 * If the World is in state RUNNING, this method causes Simulation to become PAUSED.
	 * 
	 * Internally it uses java.util.Timer and it's schedule(TimerTask task) method.
	 * 
	 * Note: In next version the more proper exception handling from this TimerTask should be implemented and it should be decoupled with Main class methods
	 * for exception handling. The exception should be caught outside of these TimerTasks.
	 * 
	 * @author Pavel Bory
	 */
	public void run(){
		if(getState() == WorldState.PAUSED || getState() == WorldState.READY){
			logger.info("Preparing to start continuous simulation");
			setState(WorldState.RUNNING);
			timer = new Timer();
			timer.schedule(new TimerTask() {
				@Override
				public void run() {
					logger.debug("Continuous simulation TimerTask.run executed");
					// TODO Timer is not executed on FX Application Thread so following way of exception handling can't be used. I will try to fix it.
					try {
						step(false);
					} catch (WorldActorOutOfGridException e) {
						//Main.HandleWorldSimulationException(e);
						setState(WorldState.TERMINATED);
					} catch (WorldActorUnsupportedOperationException e) {
						//Main.HandleWorldSimulationException(e);
						setState(WorldState.TERMINATED);
					} catch (WorldNotInStateToPerformSimulationStep e) {
						//Main.HandleWorldSimulationException(e);
						setState(WorldState.TERMINATED);
					}
				}
			}, TIMER_DELAY, TIMER_PERIOD);
			updateDebugInfo();
			logger.info("Continuous simulation started");
		} else if (getState() == WorldState.RUNNING){
			setState(WorldState.PAUSED);
			logger.info("Simulation is PAUSED");
		} else {
			logger.debug("Can't start the continuous simulation, because The World is in state: {}",getState());
		}
	}

	/**
	 * Determines whether the actor in on position in the World Grid where is placed World Item that offers him some input and if so, then returns this 
	 * input as a String.
	 * 
	 * Note: Actually only Strings can be passed to Actor as input from the World Grid Item. So in general basic types that can be simply parsed
	 * from string can be passed to actor (basic types like integer,datetime in a given format, float, strings ...)
	 */
	private String getActualItemInputForActor(Actor actor){
		if(worldSettings.getGridWorldItems()[(int) actor.getPositionX()][(int) actor.getPositionY()] != null){
			return worldSettings.getGridWorldItems()[(int) actor.getPositionX()][(int) actor.getPositionY()].getValue();
		}
		return null;
	}
	
	/**
	 * Constructs Debug Info String (StringBuilder used for optimized string creation) and passes it to the GUI
	 */
	private void updateDebugInfo(){
		logger.debug("Constructing debug info string");
		StringBuilder sBuilder = new StringBuilder();
		sBuilder.append("ACTOR:\n");
		sBuilder.append(String.format("Pos[x,y]: [%s,%s]\n", actor.getPositionX(),actor.getPositionY()));
		
		HashMap<String, Object> actorData = jrubyContainer.getActorData();
		for(String key : actorData.keySet()){
			sBuilder.append(String.format("data[%s] = %s\n",key,actorData.get(key)));
		}
		
		sBuilder.append("\n\nGRID ITEMS:\n");
		
		for(WorldItem worldItem : worldSettings.getWorldItems()){
			sBuilder.append(String.format("Pos[x,y]: [%s,%s], val: %s\n", worldItem.getPositionX(),worldItem.getPositionY(),worldItem.getValue()));
		}
		
		sBuilder.append("\n\nWORLD:\n");
		sBuilder.append(String.format("State: %s\n",getState()));
		sBuilder.append(String.format("Actor initial position X: %s\n",worldSettings.getActorInitialPositionX()));
		sBuilder.append(String.format("Actor initial position Y: %s\n",worldSettings.getActorInitialPositionY()));
		sBuilder.append(String.format("Columns count: %s\n",WORLD_COLUMNS_COUNT));
		sBuilder.append(String.format("Rows count: %s\n",WORLD_ROWS_COUNT));
		sBuilder.append(String.format("Timer interval ms: %s\n",TIMER_PERIOD));
		sBuilder.append(String.format("Timer delay ms: %s\n",TIMER_DELAY));
		
		logger.debug("Constructing debug info string - COMPLETED");
		Main.setDebugDialogText(sBuilder.toString());
	}
	
	public WorldState getState() {
		return state;
	}

	public void setState(WorldState state) {
		logger.info("Changing state from {} to {}",getState(),state);
		if(state == WorldState.RUNNING){
			if(timer != null){
				logger.info("Changing state from RUNNING -> Stopping timer");
				try{
					timer = null;
					timer.cancel();
				}catch(Exception e){
					// properly handle timer cancel threading exceptions ...
				}
			}
		}
		this.state = state;
		updateDebugInfo();
	}
}