package application.world;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javafx.scene.image.Image;
import javafx.scene.paint.Color;

/**
 * Contains the settings of the Simulation world:
 * 	Actor initial x,y coordinates withing the World Grid
 *  Image that graphically represents the Actor
 *  Two-dimensional array that contains colors definitions for World Grid cells 
 *  Two-dimensional array that contains World Items, that will be placed to World Grid cells
 * 
 * @author Pavel Bory
 *
 */
public class WorldSettings {
	private static final Logger logger = LogManager.getLogger(WorldSettings.class);
	
	private long actorInitialPositionX;
	private long actorInitialPositionY;
	private Image actorImg;
	private Color[][] gridCellColors;
	private WorldItem[][] gridWorldItems;
	
	
	public WorldSettings(long actorInitialPositionX, long actorInitialPositionY,Image actorImg, Color[][] gridCellColors,
			WorldItem[][] gridWorldItems){
		this.actorInitialPositionX = actorInitialPositionX;
		this.actorInitialPositionY = actorInitialPositionY;
		this.actorImg = actorImg;
		this.setGridCellColors(gridCellColors);
		this.setGridWorldItems(gridWorldItems);
	}
	
	/**
	* Loads settings from the given .xml File.
	* Example of xml file with settings:
	* 	<?xml version="1.0" encoding="UTF-8"?>
	* 	<world>
	* 		<!-- 	Actor initial position and image that graphically represents the Actor. 
	* 				All files referenced from this xml must be placed within the same directory
	* 				as the XML file! -->
	* 		<actor posX="0" posY="9" img="actor.png">
	* 		</actor>
	* 		<!-- Color for grid cells determined by x,y coordinates. If color for cell is not specified, the default color for cell is used -->
	*		<gridColors>
	*			<cell x="5" y="9" color="GREEN"></cell>
	*			<cell x="5" y="8" color="RED"></cell>
	*			<cell x="5" y="7" color="BLUE"></cell>
	*			<cell x="5" y="6" color="ORANGE"></cell>
	*			<cell x="6" y="9" color="VIOLET"></cell>
	*			<cell x="6" y="8" color="DARKGREEN"></cell>
	*			<cell x="6" y="7" color="DARKBLUE"></cell>
	*			<cell x="6" y="6" color="PURPLE"></cell>
	*		</gridColors>
	*		<!-- World Grid Items where each item can have custom image. 
	*			WARNING the img and val attributes are REQUIRED! There is no default image for World Grid Item,
	*			so if the img attribute is omitted and it will cause Exception during load -->
	*		<gridItems>
	*			<item x="0" y="8" img="input.png" val="6"></item>
	*			<item x="0" y="7" img="input.png" val="28"></item>
	*			<item x="0" y="6" img="input.png" val="496"></item>
	*			<item x="1" y="5" img="input.png" val="Paul"></item>
	*			<item x="1" y="4" img="input.png" val="Peter"></item>
	*		</gridItems>
	*	</world>
	*/
	public static WorldSettings load(File file) throws SAXException, IOException, ParserConfigurationException{
		logger.trace("Loading World Settings. Details: ");
		Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
			
		// Actor initial x-y coordinates for World Grid
		Node elActor = doc.getElementsByTagName("actor").item(0);
		long actorPosX = Long.parseLong(elActor.getAttributes().getNamedItem("posX").getNodeValue());
		long actorPosY = Long.parseLong(elActor.getAttributes().getNamedItem("posY").getNodeValue());
		logger.trace("Actor x,y coordinates: {},{}",actorPosX,actorPosY);
			
		// Actor image
		String actorImgFileName = elActor.getAttributes().getNamedItem("img").getNodeValue();
		Image actorImg = new Image(new File(file.getParent() + "\\" + actorImgFileName).toURI().toString());
		logger.trace("Actor image loaded from file: {}",actorImgFileName);
			
		// World Grid cell colors
		Color[][] gridCellColors = null;
		if(doc.getElementsByTagName("gridColors").getLength() > 0){
			logger.trace("Loading Grid cell colors.");
			NodeList cellColors = doc.getElementsByTagName("gridColors").item(0).getChildNodes();
			gridCellColors = new Color[World.WORLD_COLUMNS_COUNT][World.WORLD_ROWS_COUNT];
			for(int i = 0;i<cellColors.getLength();i++){
				if(cellColors.item(i).getNodeName().equals("cell")){
					int colorPosX = Integer.parseInt(cellColors.item(i).getAttributes().getNamedItem("x").getNodeValue());
					int colorPosY = Integer.parseInt(cellColors.item(i).getAttributes().getNamedItem("y").getNodeValue());
					Color color = Color.valueOf(cellColors.item(i).getAttributes().getNamedItem("color").getNodeValue());
					gridCellColors[colorPosX][colorPosY] = color;
					logger.trace("Grid color: {} loaded for cell x,y : {},{}",color,colorPosX,colorPosY);
				}
			}
		}else{
			logger.trace("No Grid cell colors to load.");
		}
			
		// World Grid Items
		WorldItem[][] gridWorldItems = null;
		if(doc.getElementsByTagName("gridItems").getLength() > 0){
			logger.trace("Loading Grid items.");
			NodeList items = doc.getElementsByTagName("gridItems").item(0).getChildNodes();
			gridWorldItems = new WorldItem[World.WORLD_COLUMNS_COUNT][World.WORLD_ROWS_COUNT];
			for(int i = 0;i<items.getLength();i++){
				if(items.item(i).getNodeName().equals("item")){
					int itemPosX = Integer.parseInt(items.item(i).getAttributes().getNamedItem("x").getNodeValue());
					int itemPosY = Integer.parseInt(items.item(i).getAttributes().getNamedItem("y").getNodeValue());
					String itemValue = items.item(i).getAttributes().getNamedItem("val").getNodeValue();
					String imgFileName = items.item(i).getAttributes().getNamedItem("img").getNodeValue();
					Image itemImg = new Image(new File(file.getParent() + "\\" + imgFileName).toURI().toString());
					WorldItem worldItem = new WorldItem(itemPosX, itemPosY, itemValue, itemImg);
					gridWorldItems[itemPosX][itemPosY] = worldItem;
					logger.trace("Grid World Item with value: {} loaded for cell x,y : {},{}. Image file name: {}",itemValue,itemPosX,itemPosY,imgFileName);
				}
			}
		}else{
			logger.trace("No Grid items to load.");
		}
		return new WorldSettings(actorPosX, actorPosY,actorImg,gridCellColors,gridWorldItems);
	}
	
	/**
	 * Returns a list of all World Grid Items placed within the World Grid.
	 */
	public List<WorldItem> getWorldItems(){
		List<WorldItem> result = new ArrayList<WorldItem>();
		for(int col = 0;col < gridWorldItems.length;col++){
			for(int row = 0; row < gridWorldItems[col].length;row++){
				if(gridWorldItems[col][row] != null){
					result.add(gridWorldItems[col][row]);
				}
			}
		}
		logger.trace("getWorldItems returning {} items",result.size());
		return result;
	}

	public long getActorInitialPositionX() {
		return actorInitialPositionX;
	}

	public void setActorInitialPositionX(long actorInitialPositionX) {
		this.actorInitialPositionX = actorInitialPositionX;
	}

	public long getActorInitialPositionY() {
		return actorInitialPositionY;
	}

	public void setActorInitialPositionY(long actorInitialPositionY) {
		this.actorInitialPositionY = actorInitialPositionY;
	}

	public Image getActorImg() {
		return actorImg;
	}

	public void setActorImg(Image actorImg) {
		this.actorImg = actorImg;
	}

	public Color[][] getGridCellColors() {
		return gridCellColors;
	}

	public void setGridCellColors(Color[][] gridCellColors) {
		this.gridCellColors = gridCellColors;
	}

	public WorldItem[][] getGridWorldItems() {
		return gridWorldItems;
	}

	public void setGridWorldItems(WorldItem[][] gridWorldItems) {
		this.gridWorldItems = gridWorldItems;
	}
}
